import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from swarms.models.openai_models import OpenAIChat


class AutoTempAgent:
    """
    AutoTemp is a tool for automatically selecting the best temperature setting for a given task.

    Flow:
        1. Generate outputs at a range of temperature settings.
        2. Evaluate each output using the default temperature setting.
        3. Select the best output based on the evaluation score.
        4. Return the best output.


    Args:
        temperature (float, optional): The default temperature setting to use. Defaults to 0.5.
        api_key (str, optional): Your OpenAI API key. Defaults to None.
        alt_temps ([type], optional): A list of alternative temperature settings to try. Defaults to None.
        auto_select (bool, optional): If True, the best temperature setting will be automatically selected. Defaults to True.
        max_workers (int, optional): The maximum number of workers to use when generating outputs. Defaults to 6.

    Returns:
        [type]: [description]

    Examples:
        >>> from swarms.demos.autotemp import AutoTemp
        >>> autotemp = AutoTemp()
        >>> autotemp.run("Generate a 10,000 word blog on mental clarity and the benefits of meditation.", "0.4,0.6,0.8,1.0,1.2,1.4")
        Best AutoTemp Output (Temp 0.4 | Score: 100.0):
        Generate a 10,000 word blog on mental clarity and the benefits of meditation.

    """

    def __init__(
        self,
        temperature: float = 0.5,
        api_key: str = None,
        alt_temps=None,
        auto_select=True,
        max_workers=6,
    ):
        self.alt_temps = alt_temps if alt_temps else [0.4, 0.6, 0.8, 1.0, 1.2, 1.4]
        self.auto_select = auto_select
        self.max_workers = max_workers
        self.temperature = temperature
        self.alt_temps = alt_temps
        self.llm = OpenAIChat(
            openai_api_key=api_key,
            temperature=temperature,
        )

    def evaluate_output(self, output: str):
        """Evaluate the output using the default temperature setting."""
        eval_prompt = f"""
            Evaluate the following output which was generated at a temperature setting of {self.temperature}. 
            Provide a precise score from 0.0 to 100.0, considering the criteria of relevance, clarity, utility, pride, and delight.

            Output to evaluate:
            ---
            {output}
            ---
            """
        score_text = self.llm(prompt=eval_prompt)
        score_match = re.search(r"\b\d+(\.\d)?\b", score_text)
        return round(float(score_match.group()), 1) if score_match else 0.0

    def run(self, task: str, temperature_string):
        """Run the AutoTemp agent."""
        temperature_list = [
            float(temp.strip()) for temp in temperature_string.split(",")
        ]
        outputs = {}
        scores = {}
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            future_to_temp = {
                executor.submit(self.llm.generate, task, temp): temp
                for temp in temperature_list
            }
            for future in as_completed(future_to_temp):
                temp = future_to_temp[future]
                output_text = future.result()
                outputs[temp] = output_text
                scores[temp] = self.evaluate_output(output_text, temp)

        if not scores:
            return "No valid outputs generated.", None

        sorted_scores = sorted(scores.items(), key=lambda item: item[1], reverse=True)
        best_temp, best_score = sorted_scores[0]
        best_output = outputs[best_temp]

        return (
            f"Best AutoTemp Output (Temp {best_temp} | Score: {best_score}):\n{best_output}"
            if self.auto_select
            else "\n".join(
                f"Temp {temp} | Score: {score}:\n{outputs[temp]}"
                for temp, score in sorted_scores
            )
        )



import logging

import torch
from torch.nn.parallel import DistributedDataParallel as DDP
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig


class WizardLLMStoryTeller:
    """
    A class for running inference on a given model.

    Attributes:
        model_id (str): The ID of the model.
        device (str): The device to run the model on (either 'cuda' or 'cpu').
        max_length (int): The maximum length of the output sequence.
        quantize (bool, optional): Whether to use quantization. Defaults to False.
        quantization_config (dict, optional): The configuration for quantization.
        verbose (bool, optional): Whether to print verbose logs. Defaults to False.
        logger (logging.Logger, optional): The logger to use. Defaults to a basic logger.

    # Usage
    ```
    from finetuning_suite import Inference

    model_id = "TheBloke/WizardLM-Uncensored-SuperCOT-StoryTelling-30B-GGUF"
    inference = Inference(model_id=model_id)

    prompt_text = "Once upon a time"
    generated_text = inference(prompt_text)
    print(generated_text)
    ```
    """

    def __init__(
        self,
        model_id: str = "TheBloke/WizardLM-Uncensored-SuperCOT-StoryTelling-30B-GGUF",
        device: str = None,
        max_length: int = 500,
        quantize: bool = False,
        quantization_config: dict = None,
        verbose=False,
        # logger=None,
        distributed=False,
        decoding=False,
    ):
        self.logger = logging.getLogger(__name__)
        self.device = (
            device if device else ("cuda" if torch.cuda.is_available() else "cpu")
        )
        self.model_id = model_id
        self.max_length = max_length
        self.verbose = verbose
        self.distributed = distributed
        self.decoding = decoding
        self.model, self.tokenizer = None, None
        # self.log = Logging()

        if self.distributed:
            assert (
                torch.cuda.device_count() > 1
            ), "You need more than 1 gpu for distributed processing"

        bnb_config = None
        if quantize:
            if not quantization_config:
                quantization_config = {
                    "load_in_4bit": True,
                    "bnb_4bit_use_double_quant": True,
                    "bnb_4bit_quant_type": "nf4",
                    "bnb_4bit_compute_dtype": torch.bfloat16,
                }
            bnb_config = BitsAndBytesConfig(**quantization_config)

        try:
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)
            self.model = AutoModelForCausalLM.from_pretrained(
                self.model_id, quantization_config=bnb_config
            )

            self.model  # .to(self.device)
        except Exception as e:
            self.logger.error(f"Failed to load the model or the tokenizer: {e}")
            raise

    def load_model(self):
        """Load the model"""
        if not self.model or not self.tokenizer:
            try:
                self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)

                bnb_config = (
                    BitsAndBytesConfig(**self.quantization_config)
                    if self.quantization_config
                    else None
                )

                self.model = AutoModelForCausalLM.from_pretrained(
                    self.model_id, quantization_config=bnb_config
                ).to(self.device)

                if self.distributed:
                    self.model = DDP(self.model)
            except Exception as error:
                self.logger.error(f"Failed to load the model or the tokenizer: {error}")
                raise

    def run(self, prompt_text: str):
        """
        Generate a response based on the prompt text.

        Args:
        - prompt_text (str): Text to prompt the model.
        - max_length (int): Maximum length of the response.

        Returns:
        - Generated text (str).
        """
        self.load_model()

        max_length = self.max_length

        try:
            inputs = self.tokenizer.encode(prompt_text, return_tensors="pt").to(
                self.device
            )

            # self.log.start()

            if self.decoding:
                with torch.no_grad():
                    for _ in range(max_length):
                        output_sequence = []

                        outputs = self.model.generate(
                            inputs, max_length=len(inputs) + 1, do_sample=True
                        )
                        output_tokens = outputs[0][-1]
                        output_sequence.append(output_tokens.item())

                        # print token in real-time
                        print(
                            self.tokenizer.decode(
                                [output_tokens], skip_special_tokens=True
                            ),
                            end="",
                            flush=True,
                        )
                        inputs = outputs
            else:
                with torch.no_grad():
                    outputs = self.model.generate(
                        inputs, max_length=max_length, do_sample=True
                    )

            del inputs
            return self.tokenizer.decode(outputs[0], skip_special_tokens=True)
        except Exception as e:
            self.logger.error(f"Failed to generate the text: {e}")
            raise

    def __call__(self, prompt_text: str):
        """
        Generate a response based on the prompt text.

        Args:
        - prompt_text (str): Text to prompt the model.
        - max_length (int): Maximum length of the response.

        Returns:
        - Generated text (str).
        """
        self.load_model()

        max_length = self.max_

        try:
            inputs = self.tokenizer.encode(prompt_text, return_tensors="pt").to(
                self.device
            )

            # self.log.start()

            if self.decoding:
                with torch.no_grad():
                    for _ in range(max_length):
                        output_sequence = []

                        outputs = self.model.generate(
                            inputs, max_length=len(inputs) + 1, do_sample=True
                        )
                        output_tokens = outputs[0][-1]
                        output_sequence.append(output_tokens.item())

                        # print token in real-time
                        print(
                            self.tokenizer.decode(
                                [output_tokens], skip_special_tokens=True
                            ),
                            end="",
                            flush=True,
                        )
                        inputs = outputs
            else:
                with torch.no_grad():
                    outputs = self.model.generate(
                        inputs, max_length=max_length, do_sample=True
                    )

            del inputs

            return self.tokenizer.decode(outputs[0], skip_special_tokens=True)
        except Exception as e:
            self.logger.error(f"Failed to generate the text: {e}")
            raise



import os
import warnings

warnings.filterwarnings("ignore", category=UserWarning)

# disable tensorflow warnings

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
from swarms.agents import *  # noqa: E402, F403
from swarms.swarms import *  # noqa: E402, F403
from swarms.structs import *  # noqa: E402, F403
from swarms.models import *  # noqa: E402, F403
from swarms.chunkers import *  # noqa: E402, F403
from swarms.workers import *  # noqa: E402, F403
